@@ -1,4 +1,3 @@
-from contextlib import contextmanager
 from typing import (
     Any,
     Collection,
@@ -6,6 +5,7 @@ from typing import (
     Generic,
     Iterable,
     Iterator,
+    List,
     Literal,
     Optional,
     Tuple,
@@ -19,13 +19,13 @@ import numpy.typing as npt
 from csnlp import Solution, wrappers
 from csnlp.util.io import SupportsDeepcopyAndPickle
 from csnlp.wrappers import Mpc
+from gymnasium import Env
 from typing_extensions import TypeAlias
 
-from mpcrl.core.callbacks import AgentCallbacks
+from mpcrl.core.callbacks import AgentCallbacks, RemovesCallbackHooksInState
 from mpcrl.core.exploration import ExplorationStrategy, NoExploration
-from mpcrl.core.random import generate_seeds
 from mpcrl.util.named import Named
-from mpcrl.util.types import GymEnvLike
+from mpcrl.util.random import generate_seeds
 
 SymType = TypeVar("SymType", cs.SX, cs.MX)
 ActType: TypeAlias = Union[npt.ArrayLike, Dict[str, npt.ArrayLike]]
@@ -39,7 +39,13 @@ def _update_dicts(sinks: Iterable[Dict], source: Dict) -> Iterator[Dict]:
         yield sink
 
 
-class Agent(Named, SupportsDeepcopyAndPickle, AgentCallbacks, Generic[SymType]):
+class Agent(
+    Named,
+    SupportsDeepcopyAndPickle,
+    AgentCallbacks,
+    RemovesCallbackHooksInState,
+    Generic[SymType],
+):
     """Simple MPC-based agent with a fixed (i.e., non-learnable) MPC controller.
 
     In this agent, the MPC controller is used as policy provider, as well as to provide
@@ -65,7 +71,6 @@ class Agent(Named, SupportsDeepcopyAndPickle, AgentCallbacks, Generic[SymType]):
         fixed_parameters: Union[
             None, Dict[str, npt.ArrayLike], Collection[Dict[str, npt.ArrayLike]]
         ] = None,
-        exploration: Optional[ExplorationStrategy] = None,
         warmstart: Literal["last", "last-successful"] = "last-successful",
         name: Optional[str] = None,
     ) -> None:
@@ -86,11 +91,6 @@ class Agent(Named, SupportsDeepcopyAndPickle, AgentCallbacks, Generic[SymType]):
             are the names of the MPC parameters and the values are their corresponding
             values. Use this to specify fixed parameters, that is, non-learnable. If
             `None`, then no fixed parameter is assumed.
-        exploration : ExplorationStrategy, optional
-            Exploration strategy for inducing exploration in the MPC policy. By default
-            `None`, in which case `NoExploration` is used in the fixed-MPC agent. While
-            learning agents need exploration to ensure convergence to a good suboptimal
-            policy, fixed agents usually do not need it.
         warmstart: 'last' or 'last-successful', optional
             The warmstart strategy for the MPC's NLP. If 'last-successful', the last
             successful solution is used to warm start the solver for the next iteration.
@@ -109,11 +109,11 @@ class Agent(Named, SupportsDeepcopyAndPickle, AgentCallbacks, Generic[SymType]):
         Named.__init__(self, name)
         SupportsDeepcopyAndPickle.__init__(self)
         AgentCallbacks.__init__(self)
+        RemovesCallbackHooksInState.__init__(self)
         self._V, self._Q = self._setup_V_and_Q(mpc)
         self._fixed_pars = fixed_parameters
-        self._exploration = NoExploration() if exploration is None else exploration
+        self._exploration: ExplorationStrategy = NoExploration()
         self._store_last_successful = warmstart == "last-successful"
-        self.reset()
 
     @property
     def unwrapped(self) -> "Agent":
@@ -148,19 +148,11 @@ class Agent(Named, SupportsDeepcopyAndPickle, AgentCallbacks, Generic[SymType]):
         """Gets the exploration strategy used within this agent."""
         return self._exploration
 
-    @contextmanager
-    def fullstate(self) -> Iterator[None]:
-        with super().fullstate(), self._Q.fullstate(), self._V.fullstate():
-            yield
-
-    @contextmanager
-    def pickleable(self) -> Iterator[None]:
-        with super().pickleable(), self._Q.pickleable(), self._V.pickleable():
-            yield
-
-    def reset(self) -> None:
-        """Resets the agent's internal variables."""
+    def reset(self, seed: Optional[int] = None) -> None:
+        """Resets the agent's internal variables and exploration's RNG."""
         self._last_solution: Optional[Solution[SymType]] = None
+        if hasattr(self.exploration, "reset"):
+            self.exploration.reset(seed)
 
     def solve_mpc(
         self,
@@ -232,7 +224,7 @@ class Agent(Named, SupportsDeepcopyAndPickle, AgentCallbacks, Generic[SymType]):
         elif isinstance(pars, dict):
             pars.update(additional_pars)
         else:  # iterable of dict
-            pars = _update_dicts(pars, additional_pars)  # type: ignore
+            pars = _update_dicts(pars, additional_pars)  # type: ignore[assignment]
         if vals0 is None and self._last_solution is not None:
             vals0 = self._last_solution.vals
 
@@ -321,16 +313,15 @@ class Agent(Named, SupportsDeepcopyAndPickle, AgentCallbacks, Generic[SymType]):
         """
         return self.solve_mpc(self._Q, state, action=action, vals0=vals0)
 
-    @staticmethod
     def evaluate(
-        agent: "Agent[SymType]",
-        env: GymEnvLike[ObsType, ActType],
+        self,
+        env: Env[ObsType, ActType],
         episodes: int,
         deterministic: bool = True,
         seed: Union[None, int, Iterable[int]] = None,
         raises: bool = True,
         env_reset_options: Optional[Dict[str, Any]] = None,
-    ) -> npt.NDArray[np.double]:
+    ) -> npt.NDArray[np.floating]:
         """Evaluates the agent in a given environment.
 
         Note: after solving `V(s)` for the current state `s`, the action is computed and
@@ -339,10 +330,8 @@ class Agent(Named, SupportsDeepcopyAndPickle, AgentCallbacks, Generic[SymType]):
 
         Parameters
         ----------
-        agent : Agent
-            The agent to evaluate.
-        env : gym env
-            A gym-like environment where to test the agent in.
+        env : Env[ObsType, ActType]
+            A gym environment where to test the agent in.
         episodes : int
             Number of evaluation episodes.
         deterministic : bool, optional
@@ -366,28 +355,28 @@ class Agent(Named, SupportsDeepcopyAndPickle, AgentCallbacks, Generic[SymType]):
             Raises if the MPC optimization solver fails and `warns_on_exception=False`.
         """
         returns = np.zeros(episodes)
-        agent.on_validation_start(env)
+        self.on_validation_start(env)
 
         for episode, current_seed in zip(range(episodes), generate_seeds(seed)):
-            agent.on_episode_start(env, episode)
-            agent.reset()
+            self.reset(seed=current_seed)
             state, _ = env.reset(seed=current_seed, options=env_reset_options)
             truncated, terminated, timestep = False, False, 0
+            self.on_episode_start(env, episode)
 
             while not (truncated or terminated):
-                action, sol = agent.state_value(state, deterministic)
+                action, sol = self.state_value(state, deterministic)
                 if not sol.success:
-                    agent.on_mpc_failure(episode, timestep, sol.status, raises)
+                    self.on_mpc_failure(episode, timestep, sol.status, raises)
 
                 state, r, truncated, terminated, _ = env.step(action)
-                agent.on_env_step(env, episode, timestep)
 
                 returns[episode] += r
                 timestep += 1
+                self.on_env_step(env, episode, timestep)
 
-            agent.on_episode_end(env, episode, returns[episode])
+            self.on_episode_end(env, episode, returns[episode])
 
-        agent.on_validation_end(env, returns)
+        self.on_validation_end(env, returns)
         return returns
 
     def _setup_V_and_Q(self, mpc: Mpc[SymType]) -> Tuple[Mpc[SymType], Mpc[SymType]]:
@@ -403,8 +392,7 @@ class Agent(Named, SupportsDeepcopyAndPickle, AgentCallbacks, Generic[SymType]):
         a0 = Q.nlp.parameter(self.init_action_parameter, (na, 1))
         perturbation = V.nlp.parameter(self.cost_perturbation_parameter, (na, 1))
 
-        actions = mpc.actions
-        u0 = cs.vertcat(*(actions[k][:, 0] for k in actions.keys()))
+        u0 = cs.vertcat(*mpc.first_actions.values())
         f = V.nlp.f
         if mpc.is_wrapped(wrappers.NlpScaling):
             f = mpc.scale(f)
@@ -419,3 +407,10 @@ class Agent(Named, SupportsDeepcopyAndPickle, AgentCallbacks, Generic[SymType]):
         """Internal utility to retrieve parameters of the MPC in order to solve it.
         `Agent` has no learnable parameter, so only fixed parameters are returned."""
         return self._fixed_pars
+
+    def __deepcopy__(self, memo: Optional[Dict[int, List[Any]]] = None) -> "Agent":
+        """Ensures that the copy has a new name."""
+        y = super().__deepcopy__(memo)
+        if hasattr(y, "name"):
+            y.name += "_copy"
+        return y
